@@ -9,12 +9,13 @@
 import six
 import abc
 import logging
+import threading
 import os
 import io
 from . import pipeline
 from azure.iot.device.common.auth import connection_string as cs
 from azure.iot.device.common.auth import sastoken as st
-
+from azure.iot.device import exceptions
 from azure.iot.device.common import auth
 from . import edge_hsm
 
@@ -31,6 +32,7 @@ def _validate_kwargs(exclude=[], **kwargs):
         "server_verification_cert",
         "proxy_options",
         "sastoken_ttl",
+        "keep_alive",
     ]
 
     for kwarg in kwargs:
@@ -46,6 +48,7 @@ def _get_config_kwargs(**kwargs):
         "cipher",
         "server_verification_cert",
         "proxy_options",
+        "keep_alive",
     ]
 
     config_kwargs = {}
@@ -64,6 +67,12 @@ def _form_sas_uri(hostname, device_id, module_id=None):
         return "{hostname}/devices/{device_id}".format(hostname=hostname, device_id=device_id)
 
 
+# Receive Type constant defs
+RECEIVE_TYPE_NONE_SET = "none_set"  # Type of receiving has not been set
+RECEIVE_TYPE_HANDLER = "handler"  # Only use handlers for receive
+RECEIVE_TYPE_API = "api"  # Only use APIs for receive
+
+
 @six.add_metaclass(abc.ABCMeta)
 class AbstractIoTHubClient(object):
     """ A superclass representing a generic IoTHub client.
@@ -79,6 +88,43 @@ class AbstractIoTHubClient(object):
         self._mqtt_pipeline = mqtt_pipeline
         self._http_pipeline = http_pipeline
 
+        self._inbox_manager = None  # this will be overriden in child class
+        self._receive_type = RECEIVE_TYPE_NONE_SET
+        self._client_lock = threading.Lock()
+
+        # Generic Client handlers
+        self._on_method_request_received = None
+        self._on_twin_desired_properties_patch_received = None
+        self._on_message_received = None
+
+    def _validate_receive_api_invoke(self):
+        """Call this function first in EVERY receive API"""
+        with self._client_lock:
+            if self._receive_type is RECEIVE_TYPE_NONE_SET:
+                # Lock the client to ONLY use receive APIs (no handlers)
+                self._receive_type = RECEIVE_TYPE_API
+            elif self._receive_type is RECEIVE_TYPE_HANDLER:
+                raise exceptions.ClientError(
+                    "Cannot use receive APIs - receive handler(s) have already been set"
+                )
+            else:
+                pass
+
+    def _validate_receive_handler_setter(self):
+        """Call this function first in EVERY handler setter"""
+        with self._client_lock:
+            if self._receive_type is RECEIVE_TYPE_NONE_SET:
+                # Lock the client to ONLY use receive handlers (no APIs)
+                self._receive_type = RECEIVE_TYPE_HANDLER
+                # Set the inbox manager to use unified msg receives
+                self._inbox_manager.use_unified_msg_mode = True
+            elif self._receive_type is RECEIVE_TYPE_API:
+                raise exceptions.ClientError(
+                    "Cannot set receive handlers - receive APIs have already been used"
+                )
+            else:
+                pass
+
     @classmethod
     def create_from_connection_string(cls, connection_string, **kwargs):
         """
@@ -100,6 +146,10 @@ class AbstractIoTHubClient(object):
         :type proxy_options: :class:`azure.iot.device.ProxyOptions`
         :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
             authentication. Default is 3600 seconds (1 hour)
+        :param int keepalive: Maximum period in seconds between communications with the
+        broker. If no other messages are being exchanged, this controls the
+        rate at which the client will send ping messages to the broker.
+        If not provided default value of 60 secs will be used.
 
         :raises: ValueError if given an invalid connection_string.
         :raises: TypeError if given an unsupported parameter.
@@ -215,6 +265,10 @@ class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
             arbitrary product info which is appended to the user agent string.
         :param proxy_options: Options for sending traffic through proxy servers.
         :type proxy_options: :class:`azure.iot.device.ProxyOptions`
+        :param int keepalive: Maximum period in seconds between communications with the
+        broker. If no other messages are being exchanged, this controls the
+        rate at which the client will send ping messages to the broker.
+        If not provided default value of 60 secs will be used.
 
         :raises: TypeError if given an unsupported parameter.
 
@@ -261,6 +315,10 @@ class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
         :type proxy_options: :class:`azure.iot.device.ProxyOptions`
         :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
             authentication. Default is 3600 seconds (1 hour)
+        :param int keepalive: Maximum period in seconds between communications with the
+        broker. If no other messages are being exchanged, this controls the
+        rate at which the client will send ping messages to the broker.
+        If not provided default value of 60 secs will be used.
 
         :raises: TypeError if given an unsupported parameter.
 
@@ -300,14 +358,6 @@ class AbstractIoTHubDeviceClient(AbstractIoTHubClient):
 
 @six.add_metaclass(abc.ABCMeta)
 class AbstractIoTHubModuleClient(AbstractIoTHubClient):
-    def __init__(self, mqtt_pipeline, http_pipeline):
-        """Initializer for a module client.
-
-        :param mqtt_pipeline: The pipeline used to connect to the IoTHub endpoint.
-        :type mqtt_pipeline: :class:`azure.iot.device.iothub.pipeline.MQTTPipeline`
-        """
-        super(AbstractIoTHubModuleClient, self).__init__(mqtt_pipeline, http_pipeline)
-
     @classmethod
     def create_from_edge_environment(cls, **kwargs):
         """
@@ -327,6 +377,10 @@ class AbstractIoTHubModuleClient(AbstractIoTHubClient):
         :type proxy_options: :class:`azure.iot.device.ProxyOptions`
         :param int sastoken_ttl: The time to live (in seconds) for the created SasToken used for
             authentication. Default is 3600 seconds (1 hour)
+        :param int keepalive: Maximum period in seconds between communications with the
+        broker. If no other messages are being exchanged, this controls the
+        rate at which the client will send ping messages to the broker.
+        If not provided default value of 60 secs will be used.
 
         :raises: OSError if the IoT Edge container is not configured correctly.
         :raises: ValueError if debug variables are invalid.
@@ -432,8 +486,8 @@ class AbstractIoTHubModuleClient(AbstractIoTHubClient):
             **config_kwargs
         )
         pipeline_configuration.method_invoke = (
-            True
-        )  # Method Invoke is allowed on modules created from edge environment
+            True  # Method Invoke is allowed on modules created from edge environment
+        )
 
         # Pipeline setup
         http_pipeline = pipeline.HTTPPipeline(pipeline_configuration)
@@ -468,6 +522,10 @@ class AbstractIoTHubModuleClient(AbstractIoTHubClient):
             arbitrary product info which is appended to the user agent string.
         :param proxy_options: Options for sending traffic through proxy servers.
         :type proxy_options: :class:`azure.iot.device.ProxyOptions`
+        :param int keepalive: Maximum period in seconds between communications with the
+        broker. If no other messages are being exchanged, this controls the
+        rate at which the client will send ping messages to the broker.
+        If not provided default value of 60 secs will be used.
 
         :raises: TypeError if given an unsupported parameter.
 
