@@ -8,7 +8,8 @@ Azure IoTHub Device SDK for Python.
 """
 
 import logging
-from azure.iot.device.common import async_adapter
+import asyncio
+from azure.iot.device.common import async_adapter, asyncio_compat
 from azure.iot.device.iothub.abstract_clients import (
     AbstractIoTHubClient,
     AbstractIoTHubDeviceClient,
@@ -20,6 +21,7 @@ from azure.iot.device.iothub.pipeline import exceptions as pipeline_exceptions
 from azure.iot.device import exceptions
 from azure.iot.device.iothub.inbox_manager import InboxManager
 from .async_inbox import AsyncClientInbox
+from . import async_handler_manager
 from azure.iot.device import constant as device_constant
 
 logger = logging.getLogger(__name__)
@@ -70,6 +72,9 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         # **kwargs.
         super().__init__(**kwargs)
         self._inbox_manager = InboxManager(inbox_type=AsyncClientInbox)
+        self._handler_manager = async_handler_manager.AsyncHandlerManager(self._inbox_manager)
+
+        # Set pipeline handlers
         self._mqtt_pipeline.on_connected = self._on_connected
         self._mqtt_pipeline.on_disconnected = self._on_disconnected
         self._mqtt_pipeline.on_method_request_received = self._inbox_manager.route_method_request
@@ -85,6 +90,46 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         self._inbox_manager.clear_all_method_requests()
         logger.info("Cleared all pending method requests due to disconnect")
 
+    async def _enable_feature(self, feature_name):
+        """Enable an Azure IoT Hub feature
+
+        :param feature_name: The name of the feature to enable.
+            See azure.iot.device.common.pipeline.constant for possible values.
+        """
+        logger.info("Enabling feature:" + feature_name + "...")
+        if not self._mqtt_pipeline.feature_enabled[feature_name]:
+            # Enable the feature if not already enabled
+            enable_feature_async = async_adapter.emulate_async(self._mqtt_pipeline.enable_feature)
+
+            callback = async_adapter.AwaitableCallback()
+            await enable_feature_async(feature_name, callback=callback)
+            await handle_result(callback)
+
+            logger.info("Successfully enabled feature:" + feature_name)
+        else:
+            # This branch shouldn't be reached, but in case it is, log it
+            logger.info("Feature ({}) already enabled - skipping".format(feature_name))
+
+    async def _disable_feature(self, feature_name):
+        """Disable an Azure IoT Hub feature
+
+        :param feature_name: The name of the feature to enable.
+            See azure.iot.device.common.pipeline.constant for possible values.
+        """
+        logger.info("Disabling feature: {}...".format(feature_name))
+        if self._mqtt_pipeline.feature_enabled[feature_name]:
+            # Disable the feature if not already disabled
+            disable_feature_async = async_adapter.emulate_async(self._mqtt_pipeline.disable_feature)
+
+            callback = async_adapter.AwaitableCallback()
+            await disable_feature_async(feature_name, callback=callback)
+            await handle_result(callback)
+
+            logger.info("Successfully disabled feature: {}".format(feature_name))
+        else:
+            # This branch shouldn't be reached, but in case it is, log it
+            logger.info("Feature ({}) already enabled - skipping".format(feature_name))
+
     async def connect(self):
         """Connects the client to an Azure IoT Hub or Azure IoT Edge Hub instance.
 
@@ -171,6 +216,8 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         :returns: MethodRequest object representing the received method request.
         :rtype: `azure.iot.device.MethodRequest`
         """
+        self._validate_receive_api_invoke()
+
         if not self._mqtt_pipeline.feature_enabled[constant.METHODS]:
             await self._enable_feature(constant.METHODS)
 
@@ -212,21 +259,6 @@ class GenericIoTHubClient(AbstractIoTHubClient):
 
         logger.info("Successfully sent method response to Hub")
 
-    async def _enable_feature(self, feature_name):
-        """Enable an Azure IoT Hub feature
-
-        :param feature_name: The name of the feature to enable.
-            See azure.iot.device.common.pipeline.constant for possible values.
-        """
-        logger.info("Enabling feature:" + feature_name + "...")
-        enable_feature_async = async_adapter.emulate_async(self._mqtt_pipeline.enable_feature)
-
-        callback = async_adapter.AwaitableCallback()
-        await enable_feature_async(feature_name, callback=callback)
-        await handle_result(callback)
-
-        logger.info("Successfully enabled feature:" + feature_name)
-
     async def get_twin(self):
         """
         Gets the device or module twin from the Azure IoT Hub or Azure IoT Edge Hub service.
@@ -299,6 +331,8 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         :returns: Twin Desired Properties patch as a JSON dict
         :rtype: dict
         """
+        self._validate_receive_api_invoke()
+
         if not self._mqtt_pipeline.feature_enabled[constant.TWIN_PATCHES]:
             await self._enable_feature(constant.TWIN_PATCHES)
         twin_patch_inbox = self._inbox_manager.get_twin_patch_inbox()
@@ -350,6 +384,26 @@ class GenericIoTHubClient(AbstractIoTHubClient):
         await handle_result(callback)
         logger.info("Successfully notified blob upload status")
 
+    # @property
+    # def on_method_request_received(self):
+    #     return self._on_method_request_received
+
+    # @on_method_request_received.setter
+    # def on_method_request_received(self, value):
+    #     self._validate_receive_handler_setter()
+    #     # Enable the feature if necessary
+    #     if value is not None and not self._mqtt_pipeline.feature_enabled[constant.METHODS]:
+    #         loop = asyncio.get_event_loop()
+    #         asyncio.ensure_future(self._enable_feature(constant.METHODS), loop=loop)
+
+    #     # Disable the feature if necessary
+    #     elif value is None and self._mqtt_pipeline.feature_enabled[constant.METHODS]:
+    #         loop = asyncio.get_event_loop()
+    #         asyncio.ensure_future(self._disable_feature(constant.METHODS), loop=loop)
+
+    #     # Set the handler on the handler manager
+    #     self._handler_manager.on_method_request_received = value
+
 
 class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
     """An asynchronous device client that connects to an Azure IoT Hub instance.
@@ -377,6 +431,8 @@ class IoTHubDeviceClient(GenericIoTHubClient, AbstractIoTHubDeviceClient):
         :returns: Message that was sent from the Azure IoT Hub.
         :rtype: :class:`azure.iot.device.Message`
         """
+        self._validate_receive_api_invoke()
+
         if not self._mqtt_pipeline.feature_enabled[constant.C2D_MSG]:
             await self._enable_feature(constant.C2D_MSG)
         c2d_inbox = self._inbox_manager.get_c2d_message_inbox()
@@ -457,6 +513,8 @@ class IoTHubModuleClient(GenericIoTHubClient, AbstractIoTHubModuleClient):
         :returns: Message that was sent to the specified input.
         :rtype: :class:`azure.iot.device.Message`
         """
+        self._validate_receive_api_invoke()
+
         if not self._mqtt_pipeline.feature_enabled[constant.INPUT_MSG]:
             await self._enable_feature(constant.INPUT_MSG)
         inbox = self._inbox_manager.get_input_message_inbox(input_name)
