@@ -4,5 +4,5 @@ This package provides pipeline for use with the Azure Provisioning Device SDK.
 
 INTERNAL USAGE ONLY
 """
-from .provisioning_pipeline import ProvisioningPipeline
+from .mqtt_pipeline import MQTTPipeline
 from .config import ProvisioningPipelineConfig
